/* $Id: exclpage.h,v 1.26 1997/07/29 12:10:42 hohmuth Exp $ 
Exclusive page for each incarnation of linux task pair
Holds pt_regs structure, and information neccessary for ipc
*/
#ifndef __EXCLPAGE_H
#define __EXCLPAGE_H

#include "config.h"

#ifndef __ASSEMBLY__
#include <l4/types.h>
#include <asm/ptrace.h> /* struct ptregs */
#include <asm/page.h>
#include "locktypes.h"
#endif

#include "task.h" /* idt types */

#define EMULIB_IPC 		(-4)  /* special value,
					 denoting msg from emulib */
#define EMULIB_SYSCALL		(-3)  /* special value denoting system call */
#define EMULIB_EXCEPTION	(-4)  /* special value denoting exception */
#define EMULIB_MSG_MASK		(0xfffffffc)
#define IS_EMULIB_MSG(msg_w0)	((msg_w0 & EMULIB_MSG_MASK) == EMULIB_IPC)

#ifndef __ASSEMBLY__
struct measure {
  unsigned long syscall_start, syscall_stop;
  unsigned long emulib_start, emulib_stop;
  unsigned long linux_server_start, linux_server_stop;
  unsigned long dispatch_start, dispatch_stop;
};

#define M_VALUES_INIT {0, 0, 0, 0, 0, 0, 0, 0}

struct exclusive_page_tag {
  struct measure m;
  char dummy[1024];
  l4_threadid_t kernel_id;
  l4_threadid_t user_id;	/* for sanity purposes */
  struct pt_regs regs;
  int exception_number;
  unsigned int error_code;
  /* local lock which prevents the signal 
     thread to generate a 0-syscall 
     when syscall is already in progress */
  simple_lock_t emu_local_lock;
  l4_idt_t user_idt;		/* space for new IDT */
  idt_descr_t user_idt_descr;
  char usignal_stack[STACK_SIZE_USIGNAL];
  char uemulib_stack[STACK_SIZE_UEMULIB];
};

#define EX_PAGE_INITIALIZER { M_VALUES_INIT, {0, }, L4_NIL_ID, L4_NIL_ID, \
			      PTREGS_INITIALIZER, 0, 0, \
			      0, {{0,0},}, {0,0}, {"SIGNALSTACK"}, {"EMUSTACK" }};

#define exclusive_page_ptr ((struct exclusive_page_tag *) EXCLUSIVE_PAGE_ADDRESS)

#endif /* __ASSEMBLY__ */

#define ZERO_SYSCALL_NUMBER 0xef

/* offsets for exclusive page members */
#define EMULIB_START_ADDRESS (EXCLUSIVE_PAGE_ADDRESS + 2*4)
#define EMULIB_STOP_ADDRESS (EXCLUSIVE_PAGE_ADDRESS + 3*4)

#define KERNEL_ID_ADDRESS  (EXCLUSIVE_PAGE_ADDRESS + 8*4 + 1024)
#define PTREGS_ADDRESS     (KERNEL_ID_ADDRESS+8+8)
#define EXCEPTION_NO_ADDRESS (PTREGS_ADDRESS+PTREGS_SIZE)
#define ERROR_CODE_ADDRESS (EXCEPTION_NO_ADDRESS+4)
#define EMU_LOCAL_LOCK_ADDRESS (ERROR_CODE_ADDRESS+4)
#define IDT_ADDRESS        (EMU_LOCAL_LOCK_ADDRESS+4)
#define IDT_DESCR_ADDRESS  (IDT_ADDRESS+IDT_SIZE)
#define USIGNAL_STACK_INIT (IDT_DESCR_ADDRESS+IDT_DESCR_SIZE \
			    +STACK_SIZE_USIGNAL)
#define UEMULIB_STACK_INIT (USIGNAL_STACK_INIT+STACK_SIZE_UEMULIB)

#define OFFS_PTREGS (PTREGS_ADDRESS - EXCLUSIVE_PAGE_ADDRESS)
#endif

